/*
 * filter.h
 *
 *  Created on: Apr 26, 2016
 *      Author: vlegrand
 */
#ifndef FILTER_HPP_
#define FILTER_HPP_
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "CountMinSketch.hpp"
#include "ReadProcessor.h"
#include "read_utils.h"


static const unsigned short ushortmask=65535;
static const unsigned char ubytemask=255;

typedef CountMinSketch<unsigned char> ByteCountMinSketch;
typedef CountMinSketch<unsigned short> ShortCountMinSketch;



// created this class to hide implementation detail (byte versys short) from main program.
class Filter {
    ByteCountMinSketch * pByteCMS;
    ShortCountMinSketch * pShortCMS;

    FasqQualThreshold qual_thres;
    long nb_bytes_before_fill_CMS,nb_bytes_after_fill_CMS;

    void getRSS(int before_fill=0);


    template <typename T> void underlyingfillCMS(FqBaseBackend* map_id_backend[],int nb_be, int k, srp* io_sr, CountMinSketch<T>* cms);
    template <typename T> void underlyinglowFilterCMS(FqBaseBackend* map_id_backend[],int nb_be,int k,srp* io_sr, CountMinSketch<T>* pcms);

public:

    Filter(const CMSparams& parms,const FasqQualThreshold& the_qual_thres) {
        pByteCMS=NULL;
        pShortCMS=NULL;
        getRSS();
        if (parms.kappa<ubytemask) pByteCMS=new ByteCountMinSketch(parms);
        else pShortCMS=new ShortCountMinSketch(parms);
        qual_thres=the_qual_thres;

        nb_bytes_after_fill_CMS=0;
    }

    void fillCMS(FqBaseBackend* map_id_backend[],int nb_be,int k,srp* io_sr);
    void lowFilterCMS(FqBaseBackend* map_id_backend[],int nb_be,int k, srp* io_sr);
    unsigned long getApproxNbDistinctKMers();
    unsigned long getSize();
};

template <typename T> void Filter::underlyingfillCMS(FqBaseBackend* map_id_backend[],int nb_be, int k, srp* io_sr, CountMinSketch<T>* cms) {
    int ret;
    srp::reverse_iterator rit;
    i_dim::iterator it_offs;
    k_dim::iterator it_struct;
    ReadProcessor read_p(k);
    T_read_numericValues nbrKmerDecompo;

    // 1rst step, open files before fetching reads in them
    int i;
    for (i=0;i<nb_be;i++){
        map_id_backend[i]->openInputFile();
    }

    for (rit=io_sr->rbegin(); rit!=io_sr->rend(); ++rit) { //process map in reverse order (by decreasing scores).
        for (it_offs=rit->second.begin();it_offs!=rit->second.end();it_offs++) {
            unsigned long j=it_offs->first;
            for (it_struct=it_offs->second.begin();it_struct!=it_offs->second.end();it_struct++) {
                // read dna string corresponding to fast qrecord
                DnaSeqStr a_seqs[2];
                init_DnaSeqStr(&a_seqs[0]);
                init_DnaSeqStr(&a_seqs[1]);
                getDNASeqstr(map_id_backend,*it_struct,j,a_seqs,qual_thres.nucl_score_threshold);
                // decompose dna string into k-mers and turn k_mers into numbers.
                decomposeReadInKMerNums(read_p, nbrKmerDecompo,k,cms->arePEFilteredAsSingle(),a_seqs);
                // insert k-mers into CMS if overall read satisfies kappa condition.
                ret=cms->addRead(nbrKmerDecompo);
                if (!ret) {
                    // read is rejected, update rpos structure accordingly.
                    it_struct->fileid=0;
                }
                nbrKmerDecompo.single_or_PE_val.clear();
            }
        }
    }
    // last step, close fq files.
    for (i=0;i<nb_be;i++){
        map_id_backend[i]->closeInputFile();
    }
}

template <typename T> void Filter::underlyinglowFilterCMS(FqBaseBackend* map_id_backend[],int nb_be,int k,srp* io_sr, CountMinSketch<T>* cms) { // possible refactor: get k from CMS.
    srp::iterator it;
    i_dim::iterator it_offs;
    k_dim::iterator it_struct;
    int ret;
    ReadProcessor read_p(k);
    T_read_numericValues nbrKmerDecompo;

    // 1rst step, open files before fetching reads in them
    int i;
    for (i=0;i<nb_be;i++){
        map_id_backend[i]->openInputFile();
    }

    for (it=io_sr->begin(); it!=io_sr->end(); ++it) {
        for (it_offs=it->second.begin();it_offs!=it->second.end();it_offs++) {
            unsigned int j=it_offs->first;
            for (it_struct=it_offs->second.begin();it_struct!=it_offs->second.end();it_struct++) {
                // read dna string corresponding to fastq record
                if (it_struct->fileid) {
                DnaSeqStr a_seqs[2];
                init_DnaSeqStr(&a_seqs[0]);
                init_DnaSeqStr(&a_seqs[1]);
                getDNASeqstr(map_id_backend,*it_struct,j,a_seqs);
                // decompose dna string into k-mers and turn k_mers into numbers.
                decomposeReadInKMerNums(read_p, nbrKmerDecompo,k,cms->arePEFilteredAsSingle(),a_seqs);
                ret=cms->isBeneathMinKappa(nbrKmerDecompo);
                if (ret) it_struct->fileid=0;
                nbrKmerDecompo.single_or_PE_val.clear();
                }
            }
        }
    }
    for (i=0;i<nb_be;i++){
        map_id_backend[i]->closeInputFile();
    }
}

void Filter::fillCMS(FqBaseBackend* map_id_backend[],int nb_be,int k,srp* io_sr) {
    if (pByteCMS!=NULL) underlyingfillCMS<unsigned char>(map_id_backend,nb_be, k, io_sr,pByteCMS);
    else underlyingfillCMS<unsigned short>(map_id_backend,nb_be, k, io_sr,pShortCMS);
    getRSS(1);
}

void Filter::lowFilterCMS(FqBaseBackend* map_id_backend[],int nb_be,int k, srp* io_sr) {
    if (pByteCMS!=NULL) underlyinglowFilterCMS<unsigned char>(map_id_backend,nb_be, k, io_sr,pByteCMS);
    else underlyinglowFilterCMS<unsigned short>(map_id_backend,nb_be, k, io_sr,pShortCMS);
}

unsigned long Filter::getApproxNbDistinctKMers() {
    if (pByteCMS!=NULL) return pByteCMS->getNbDistinctKMers();
    else return pShortCMS->getNbDistinctKMers();
}

void Filter::getRSS(int before_fill) {
    struct rusage usage;
    int res2=getrusage(RUSAGE_SELF,&usage);
    if (res2==-1) err(errno,"cannot get resource usage.");
    if (before_fill==0) nb_bytes_before_fill_CMS=usage.ru_maxrss;
    else nb_bytes_after_fill_CMS=usage.ru_maxrss;
}

unsigned long Filter::getSize() {
    unsigned long cms_size=nb_bytes_after_fill_CMS-nb_bytes_before_fill_CMS;
    // cms_size/=8; ru_maxrss seems to be already in Bytes.
    return cms_size;
}

#endif
