/*
 * main_utils.cpp
 *
 *  Created on: May 9, 2016
 *      Author: vlegrand
 */
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/stat.h>
#include <limits.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <cmath>
#include "main_utils.h"

using namespace std;

/*
 * Use this function to get the amount of RAM in GB on the machine (used or not).
 * Aim is to make sure that ROCK can run on the current machine with the requirements that the user gave.
 */
unsigned long getNodePhysMemory() {
    uint64_t mem = 0;

#if defined(_SC_PHYS_PAGES) && defined(_SC_PAGESIZE)
    mem = sysconf(_SC_PHYS_PAGES);
    mem *= sysconf(_SC_PAGESIZE);
    mem /=  1 * 1024 * 1024 * 1024;
#elif __APPLE__
    size_t len = sizeof(mem);
    int ret=sysctlbyname("hw.memsize", (void*) &mem, &len, NULL,0);
    if (ret==-1) {
        cout<<"Couldn't determine how much RAM is usable on your system."<<errno<<endl;
	mem=0;
    }
    mem /=  1 * 1024 * 1024 * 1024;
#else
    cout<<"ROCK has not been tested on this operating system, please contact author."<<endl;
#endif

    return (unsigned long)mem;
}


/*
 * Given a filename (including full path), determine if parent directory exists.
 * If is doesn't, display error message and exit.
 */
std::string checkDirExists(const string o_file) {
    string parent_dir=".";
    std::size_t o_found = o_file.find_last_of("/");
    if (o_found!=string::npos) {
        parent_dir=o_file.substr(0,o_found);
        // cout<<parent_dir<<endl;
        struct stat info;
        if (stat(parent_dir.c_str(),&info)!=0) {
            cout<<"parent directory for output files: "<<parent_dir<<" doesn't exist."<<endl;
            exit(EXIT_FAILURE);
        }
    }
    return parent_dir;
}


/*
 * Minimize lambda for a given number of k-mers.
 * We want p<=0.01,
 * so choose smallest lambda so that [1-(1-1/m)exp n] exp lambda<=0.01
 */
int getBestLambdaForN(const unsigned long& nb_k_mers,int lambda_max) {
    int lambda=2;
    int min_lambda=lambda;
    if (lambda_max==0) { // no upper bound specified by the user via the -g option
        lambda_max=500;
    }
    long double tmp=1.0/INT_MAX;
    tmp=1.0-tmp;
    tmp=pow(tmp,nb_k_mers);
    tmp=1.0-tmp;
    while (lambda<=lambda_max) {
        min_lambda=lambda;
        long double p=pow(tmp,lambda);
        if (p<=0.01) break;
        lambda+=1;
    }
    return min_lambda;
}

/*
 * Compute collision probability
 */
float getCollisionProba(const unsigned long& nb_k_mers,const int& lambda) {
    long double tmp=1.0/INT_MAX;
    tmp=1.0-tmp;
    tmp=pow(tmp,nb_k_mers);
    tmp=1.0-tmp;
    long double p=pow(tmp,lambda);
    if (p<0.0001) p=0;
    // rounding
    p=p*1000+0.5;
    return floor(p)/1000;
    //return p;
}

void usage(int status) {
  cout<<"usage: rock [options] [args]"<<endl<<endl;
  cout<<"    -i <file>      .... file name containing the list of the FASTQ files to process."<<endl;
  cout<<"    -o <file>      .... file name containing the list of the output FASTQ file names."<<endl;
  cout<<"    -h             .... Print this message and exit."<<endl;
  cout<<"    -k <int>       .... k-mer size. (default 30)."<<endl;
  cout<<"    -c <int>       .... lower coverage threshold (default: 5)."<<endl;
  cout<<"    -C <int>       .... upper coverage threshold (default: 70; max: 65535)."<<endl;
  cout<<"    -l <int>       .... size of the count min sketch (default: at most 4, depending on the available RAM)"<<endl;
  cout<<"    -n <int>       .... (expected) number of distinct k-mers within the processed reads."<<endl;
  cout<<"    -g <int>       .... maximum RAM to use (in Gb) for the CMS."<<endl;
  cout<<"    -p             .... treat PE as single. Default is NO."<<endl;
  cout<<"    -q <int>       .... minimum quality threshold for nucleotides. Default is 10."<<endl;
  cout<<"    -m <int>       .... minimum of valid k-mer threshold for reads. Default is 1."<<endl;
  cout<<"    -v             .... verbose"<<endl;
  exit(status); }

