/*
 * unit_test_cms.cpp
 *
 *  Created on: Feb 9, 2016
 *      Author: vlegrand
 *
 *  Here, gather the tests for the ROCK's main component: the CMS.
 */
#include <iostream>
#include <assert.h>
#include <climits>

using namespace std;
#include "CountMinSketch.hpp"
#include "main_utils.h"


void test_CMS(int lambda,int kappa,int kappa_prime) {
    CountMinSketch<unsigned char> cms(lambda,kappa,kappa_prime);
    int i;
    cout<<"size of the CMS component: "<<sizeof(cms)<<endl;
    int num=100*lambda;
    int rej_expected=0;
    int ret;
    for (i=0;i<num;i++) {
        cms.addKMer(num);
    }
    // Now, check that our k-mer is present in the CMS.
    int min=cms.getEstimatedNbOcc(num);
    std::cout<<"min="<<min<<endl;
    assert((min==num) || (min==num-get_mask<unsigned char>::value-1)); // addKmer doesn't check kappa.
    
    unsigned long nb_distinct_k_mers=cms.getNbDistinctKMers();
    assert(nb_distinct_k_mers==1);

    // mimic a read (as given by part 2 output).
    T_read_numericValues read_values;
    for (i=1;i<=400;i++) {
        read_values.single_or_PE_val.push_back(i*1000-1);
    }
    read_values.idx_start_PE2=0;
    ret=cms.addRead(read_values); // read should be accepted since none of the values that it contains is equal to those I inserted previously.
    assert(ret!=rej_expected);

    nb_distinct_k_mers=cms.getNbDistinctKMers();
    assert(nb_distinct_k_mers==401);

    // mimic a vector that contains a lot of k_mers already present in the CMS and check that it is rejected.
    T_read_numericValues read_values2;
    read_values2.idx_start_PE2=0;
    for (i=1;i<=199;i++) {
        read_values2.single_or_PE_val.push_back(i*1000-1);
    }
    for (i=200;i<=400;i++) {
        read_values2.single_or_PE_val.push_back(num);
    }
    ret=cms.addRead(read_values2);
    assert(ret==rej_expected);

    nb_distinct_k_mers=cms.getNbDistinctKMers();
    assert(nb_distinct_k_mers==401);

    // test that reads with median cov<kappa_prime are rejected.
    ret=cms.isBeneathMinKappa(read_values);
    assert(ret>0);
}



void testCMS_with_PE_NotasSingle(int lambda,int kappa,int kappa_prime) {
    int filter_PE_as_single=0;
    CountMinSketch<unsigned char>* cms=new CountMinSketch<unsigned char>(lambda,kappa,kappa_prime,filter_PE_as_single);
    int i;
    cout<<"size of the CMS component: "<<sizeof(cms)<<endl;
    int num=200;
    //int rej_expected=0;
    int ret;
    for (i=0;i<kappa;i++) {
        cms->addKMer(num);
    }
    int min=cms->getEstimatedNbOcc(num);
    assert(min==kappa);
    // mimic a PE read  (as given by part 2 output).
    T_read_numericValues read_values;

    // mimic a PE read in which PE1 contains k-mer that appear rarely in the CMS and PE2 contains k_mers that already appear kappa times in the CMS.
    for (i=1;i<=200;i++) {
        read_values.single_or_PE_val.push_back(i*1000-1);
    }
    read_values.idx_start_PE2=200;
    for (i=1;i<=239;i++) {
        read_values.single_or_PE_val.push_back(num);
    }
    read_values.single_or_PE_val.push_back(50);
    read_values.single_or_PE_val.push_back(55);
    ret=cms->addRead(read_values); // read should be accepted since PE1 is nor over covered.
    assert(ret>0);

    ret=cms->isBeneathMinKappa(read_values); // read should not be considered under covered since only PE1 is under covered.
    assert(ret==0);

    // Now, change PE1 a little and insert it kappa times in the CMS
    for (int j=1;j<kappa;j++) {
        read_values.single_or_PE_val[2]=j;
        ret=cms->addRead(read_values);
        if (j<kappa-1) assert(ret>0); // read should be accepted as long as we haven't inserted PE1 kappa times.
        else assert(ret>0); //otherwise it is rejected.
    }

    ret=cms->isBeneathMinKappa(read_values); // read should not be considered under covered since PE1 was inserted kappa times.
    assert(ret==0);

    // mimic a PE read  (as given by part 2 output).
    T_read_numericValues read_values2;

    // mimic a PE read that will be inserted in the CMS kappa_prime+1 times to check low filter.
    for (i=1;i<=150;i++) {
        read_values2.single_or_PE_val.push_back(i*1500-1);
    }
    read_values2.idx_start_PE2=150;
    for (i=1;i<=151;i++) {
        read_values2.single_or_PE_val.push_back(i*1600-1);
    }
    for (i=0;i<kappa_prime+1;i++) {
        ret=cms->addRead(read_values2);
        assert(ret>0);
    }

    ret=cms->isBeneathMinKappa(read_values2);
    assert(ret==0);

    //mimic a read for which PE1 is under kappa_prime and PE2 too. Such a read is expected to be removed by low filter
    T_read_numericValues read_values3;

    for (i=1;i<=200;i++) {
        read_values3.single_or_PE_val.push_back(i);
    }
    read_values3.idx_start_PE2=201;
    for (i=1;i<=205;i++) {
        read_values3.single_or_PE_val.push_back(i);
    }
    ret=cms->addRead(read_values3);
    assert(ret>0);

    ret=cms->isBeneathMinKappa(read_values3);
    assert(ret>0);

    //mimic a read for which PE1 is over kappa_prime but not PE2
    T_read_numericValues read_values4;
    for (i=201;i<=399;i++) {
        read_values4.single_or_PE_val.push_back(200);
    }
    read_values4.idx_start_PE2=199;
    for (i=201;i<=399;i++) {
        read_values4.single_or_PE_val.push_back(i);
    }
    ret=cms->addRead(read_values4);
    assert(ret>0);

    ret=cms->isBeneathMinKappa(read_values4);
    assert(ret==0);

    //mimic a read for which PE2 is over kappa_prime but not PE1
    T_read_numericValues read_values5;
    for (i=201;i<=399;i++) {
        read_values5.single_or_PE_val.push_back(i);
    }
    read_values5.idx_start_PE2=199;
    for (i=201;i<=399;i++) {
        read_values5.single_or_PE_val.push_back(200);
    }
    ret=cms->addRead(read_values5);
    assert(ret>0);

    ret=cms->isBeneathMinKappa(read_values5);
    assert(ret==0);

    delete cms;
}

/*
 * TODO: move this somewhere else.
 * I use this test because it is in the calculation of the median that we spend the most time (approx: 50%) . I want to know what in it takes the longest to see if I can improve.
 */
/*
#include <time.h>
#include <stdlib.h>
#include <err.h>
#include "Filter.hpp"
void profiling_CMS_fill() {
    ByteCountMinSketch * pByteCMS;
    CMSparams parms;
    parms.lambda=4;
    parms.filter_PE_as_single=0;
    parms.kappa=50;
    parms.kappa_prime=20;
    parms.filter_size=getNodePhysMemory()/100.0*90-1;
    pByteCMS=new ByteCountMinSketch(parms);
    // mimic SRR dataset: approx 3 million PE of 2*250 nucleotides (2*281 k-mers if k=30).
    T_read_numericValues nbrKmerDecompo;
    nbrKmerDecompo.idx_start_PE2=281;
    srand (time(NULL));
    nbrKmerDecompo.single_or_PE_val.reserve(502);

    for (int i=0; i<3000000;i++) {
        //printf("x=%d\n",x);
        for (int j=0;j<502;j++) {
            int x=rand() % 1000000 + 1;
            nbrKmerDecompo.single_or_PE_val.push_back(x);
        }
        int keep_r=pByteCMS->addRead(nbrKmerDecompo);
        nbrKmerDecompo.single_or_PE_val.clear();
    }
    delete(pByteCMS);
}
*/

int main(int argc, char **argv) {
    int lambda=2;
    int kappa=50;
    int kappa_prime=20;
    cout<<"INT_MAX="<<INT_MAX<<endl;
    cout<<"UINT_MAX="<<UINT_MAX<<endl;
    cout<<"LONG_MAX="<<LONG_MAX<<endl;
    cout<<"ULONG_MAX="<<ULONG_MAX<<endl;
    cout<<"ULONG_MAX/UINT_MAX="<<ULONG_MAX/UINT_MAX<<endl; // max value for j
    cout<<"ULONG_MAX%UINT_MAX="<<(ULONG_MAX-1)%UINT_MAX<<endl;
    cout<<"sizeof(short)="<<sizeof(short)<<endl;
    cout<<"sizeof(int)="<<sizeof(int)<<endl;
    cout<<"sizeof(long)="<<sizeof(long)<<endl;

    cout<<"checking that your machine has enough RAM to run test"<<endl;
    unsigned long ram=getNodePhysMemory();
    cout<<"machine has: "<<ram<<" GB of RAM"<<endl;

    if (ram<5) {
        cout<<"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"<<endl;
        cout<<" WARNING: skipping this test, it requires at least 4 gigabytes of RAM to run."<<endl;
        cout<<"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"<<endl;
        return -1;
    }
    

    cout<<"testing CMS with lambda="<<lambda<<endl;
    test_CMS(lambda,kappa,kappa_prime); // Finally using C arrays (maps implied storing hash keys : 4 Bytes per k_mer overhead) but each array is of size INT_MAX...
    cout<<"testing CMS with PE not as single"<<endl;
    testCMS_with_PE_NotasSingle(lambda,kappa,kappa_prime);
    cout<<"done"<<endl;
    
    // profiling_CMS_fill();
    return 0;
}
